# Generated by Django 2.2.20 on 2021-09-20 21:29

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("course_flow", "0069_auto_20210909_2140"),
    ]

    operations = [
        migrations.AddField(
            model_name="workflow",
            name="time_general_hours",
            field=models.PositiveIntegerField(default=0, null=True),
        ),
        migrations.AddField(
            model_name="workflow",
            name="time_specific_hours",
            field=models.PositiveIntegerField(default=0, null=True),
        ),
        migrations.AlterField(
            model_name="column",
            name="comments",
            field=models.ManyToManyField(
                blank=True, related_name="column", to="course_flow.Comment"
            ),
        ),
        migrations.AlterField(
            model_name="node",
            name="comments",
            field=models.ManyToManyField(
                blank=True, related_name="node", to="course_flow.Comment"
            ),
        ),
        migrations.AlterField(
            model_name="outcome",
            name="comments",
            field=models.ManyToManyField(
                blank=True, related_name="outcome", to="course_flow.Comment"
            ),
        ),
        migrations.AlterField(
            model_name="week",
            name="comments",
            field=models.ManyToManyField(
                blank=True, related_name="week", to="course_flow.Comment"
            ),
        ),
        migrations.AlterField(
            model_name="workflow",
            name="ponderation_individual",
            field=models.PositiveIntegerField(default=0, null=True),
        ),
        migrations.AlterField(
            model_name="workflow",
            name="ponderation_practical",
            field=models.PositiveIntegerField(default=0, null=True),
        ),
        migrations.AlterField(
            model_name="workflow",
            name="ponderation_theory",
            field=models.PositiveIntegerField(default=0, null=True),
        ),
    ]



import re

from course_flow import redux_actions as actions

from .models import (
    Node,
    NodeWeek,
    Outcome,
    OutcomeOutcome,
    OutcomeWorkflow,
    Week,
    WeekWorkflow,
)
from .serializers import (
    NodeSerializerShallow,
    NodeWeekSerializerShallow,
    OutcomeOutcomeSerializerShallow,
    OutcomeSerializerShallow,
    OutcomeWorkflowSerializerShallow,
    WeekSerializerShallow,
    WeekWorkflowSerializerShallow,
)
from .utils import save_serializer


def add_to_outcome_parent_with_depth(to_add, base_outcome, required_depth):
    if base_outcome.depth == required_depth:
        return OutcomeOutcome.objects.create(
            child=to_add,
            parent=base_outcome,
            rank=base_outcome.children.all().count(),
        )
    else:
        return add_to_outcome_parent_with_depth(
            to_add, base_outcome.parent_outcomes.first(), required_depth
        )


def import_outcomes(df, workflow, user):

    last_outcome = None
    for index, row in df.iterrows():
        code = str(row["code"])
        if code is not None and isinstance(code, str) and code.find(".") >= 0:
            code = re.search("([^.]+$)", code).group(0)
        title = row["title"]
        description = row["description"]

        try:
            depth = int(row["depth"])
        except ValueError:
            depth = 0
        if last_outcome is None:
            depth = 0
        elif depth > 2:
            depth = last_outcome.depth
        elif last_outcome.depth + 1 < depth:
            depth = last_outcome.depth + 1

        outcome = Outcome.objects.create(author=user, depth=depth)
        serializer = OutcomeSerializerShallow(
            outcome,
            data={"title": title, "description": description, "code": code},
            partial=True,
        )
        save_serializer(serializer)
        if depth == 0:
            outcomeworkflow = OutcomeWorkflow.objects.create(
                workflow=workflow,
                outcome=outcome,
                rank=workflow.outcomes.all().count(),
            )
            response_data = {
                "new_model": OutcomeSerializerShallow(outcome).data,
                "new_through": OutcomeWorkflowSerializerShallow(
                    outcomeworkflow
                ).data,
                "parentID": workflow.id,
            }
            actions.dispatch_wf(
                workflow, actions.newOutcomeAction(response_data)
            )
            actions.dispatch_to_parent_wf(
                workflow, actions.newOutcomeAction(response_data)
            )
        else:
            outcomeoutcome = add_to_outcome_parent_with_depth(
                outcome, last_outcome, depth - 1
            )

            new_model_serialized = OutcomeSerializerShallow(outcome).data
            new_through_serialized = OutcomeOutcomeSerializerShallow(
                outcomeoutcome
            ).data
            response_data = {
                "new_model": new_model_serialized,
                "new_through": new_through_serialized,
                "parentID": outcomeoutcome.parent.id,
            }
            actions.dispatch_wf(
                workflow,
                actions.insertChildAction(response_data, "outcome"),
            )
            actions.dispatch_to_parent_wf(
                workflow,
                actions.insertChildAction(response_data, "outcome"),
            )
        try:
            if isinstance(code, str) and code.isnumeric():
                if outcome.depth == 0:
                    rank = OutcomeWorkflow.objects.get(outcome=outcome).rank
                else:
                    rank = OutcomeOutcome.objects.get(child=outcome).rank
                if int(code) == rank or int(code) == rank + 1:
                    outcome.code = ""
                    outcome.save()
        except ValueError:
            pass

        last_outcome = outcome


def import_nodes(df, workflow, user):
    week_rank = -1
    week = None
    for index, row in df.iterrows():
        type = row.get("type", "node")
        if type == "week":
            week_rank += 1
            created = False
            try:
                week = workflow.weeks.filter(deleted=False).order_by(
                    "weekworkflow__rank"
                )[week_rank]
                weekworkflow = WeekWorkflow.objects.get(week=week)
            except IndexError:
                week = Week.objects.create(
                    author=user,
                    week_type=workflow.get_subclass().WORKFLOW_TYPE,
                )
                weekworkflow = WeekWorkflow.objects.create(
                    week=week,
                    workflow=workflow,
                    rank=workflow.weeks.all().count(),
                )
                created = True
            title = row.get("title", "")
            description = row.get("description", "")
            data = {}
            if title is not None and title != "":
                data["title"] = title
            if description is not None and description != "":
                data["description"] = description
            serializer = WeekSerializerShallow(
                week,
                data=data,
                partial=True,
            )
            save_serializer(serializer)
            if created:
                response_data = {
                    "new_model": WeekSerializerShallow(week).data,
                    "new_through": WeekWorkflowSerializerShallow(
                        weekworkflow
                    ).data,
                    "parentID": workflow.id,
                }
                actions.dispatch_wf(
                    workflow, actions.insertBelowAction(response_data, "week")
                )
            else:
                actions.dispatch_wf(
                    workflow,
                    actions.changeField(week.id, "week", data, False),
                )

        elif type == "node":
            if week is None:
                week = (
                    workflow.weeks.filter(deleted=False)
                    .order_by("weekworkflow__rank")
                    .first()
                )
            try:
                column_rank = int(row.get("column_order", 0))
                column = workflow.columns.filter(deleted=False).order_by(
                    "columnworkflow__rank"
                )[column_rank]
            except (IndexError, ValueError):
                column = (
                    workflow.columns.filter(deleted=False)
                    .order_by("columnworkflow__rank")
                    .first()
                )

            node = Node.objects.create(
                author=user,
                column=column,
                has_autolink=False,
                node_type=workflow.get_subclass().WORKFLOW_TYPE,
            )
            title = row.get("title", "")
            description = row.get("description", "")
            data = {}
            if title is not None and title != "":
                data["title"] = title
            if description is not None and description != "":
                data["description"] = description
            serializer = NodeSerializerShallow(node, data=data, partial=True)
            save_serializer(serializer)
            nodeweek = NodeWeek.objects.create(
                node=node, week=week, rank=week.nodes.all().count()
            )
            response_data = {
                "new_model": NodeSerializerShallow(node).data,
                "new_through": NodeWeekSerializerShallow(nodeweek).data,
                "parentID": week.id,
            }
            actions.dispatch_wf(
                workflow, actions.insertBelowAction(response_data, "node")
            )



# Generated by Django 2.2.16 on 2021-03-31 17:48

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("contenttypes", "0002_remove_content_type_name"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("course_flow", "0053_auto_20210318_2131"),
    ]

    operations = [
        migrations.CreateModel(
            name="Favourite",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("object_id", models.PositiveIntegerField()),
                (
                    "content_type",
                    models.ForeignKey(
                        limit_choices_to={
                            "model__in": [
                                "project",
                                "activity",
                                "course",
                                "program",
                                "outcome",
                            ]
                        },
                        on_delete=django.db.models.deletion.CASCADE,
                        to="contenttypes.ContentType",
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="ObjectPermission",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("object_id", models.PositiveIntegerField()),
                (
                    "permission_type",
                    models.PositiveIntegerField(
                        choices=[(0, "None"), (1, "View"), (2, "Edit")],
                        default=0,
                    ),
                ),
                (
                    "content_type",
                    models.ForeignKey(
                        limit_choices_to={
                            "model__in": [
                                "project",
                                "activity",
                                "course",
                                "program",
                                "outcome",
                            ]
                        },
                        on_delete=django.db.models.deletion.CASCADE,
                        to="contenttypes.ContentType",
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
        ),
        migrations.RemoveField(
            model_name="coursefavourite",
            name="course",
        ),
        migrations.RemoveField(
            model_name="coursefavourite",
            name="user",
        ),
        migrations.RemoveField(
            model_name="outcomefavourite",
            name="outcome",
        ),
        migrations.RemoveField(
            model_name="outcomefavourite",
            name="user",
        ),
        migrations.RemoveField(
            model_name="programfavourite",
            name="program",
        ),
        migrations.RemoveField(
            model_name="programfavourite",
            name="user",
        ),
        migrations.RemoveField(
            model_name="projectfavourite",
            name="project",
        ),
        migrations.RemoveField(
            model_name="projectfavourite",
            name="user",
        ),
        migrations.DeleteModel(
            name="ActivityFavourite",
        ),
        migrations.DeleteModel(
            name="CourseFavourite",
        ),
        migrations.DeleteModel(
            name="OutcomeFavourite",
        ),
        migrations.DeleteModel(
            name="ProgramFavourite",
        ),
        migrations.DeleteModel(
            name="ProjectFavourite",
        ),
    ]



# Generated by Django 2.2.20 on 2021-09-07 17:13

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("course_flow", "0067_customterm_translation_plural"),
    ]

    operations = [
        migrations.AddField(
            model_name="workflow",
            name="code",
            field=models.CharField(blank=True, max_length=50, null=True),
        ),
    ]



from typing import List

import pytest
from django.contrib.auth.models import User

from course_flow.lti import generate_password
from course_flow.models import Course, Node


@pytest.fixture(autouse=True)
def enable_db_access_for_all_tests(db):
    pass


@pytest.fixture
def user() -> User:
    return User.objects.create_user(
        username="test", password=generate_password("test")
    )


@pytest.fixture
def users() -> List[User]:
    return [
        User.objects.create_user(
            username=f"test{i}", password=generate_password(f"test{i}")
        )
        for i in range(3)
    ]


@pytest.fixture
def courses(users):
    return [
        Course.objects.create(
            title=f"test{i}", description="test", author=users[i % len(users)]
        )
        for i in range(10)
    ]


@pytest.fixture
def node(user) -> Node:
    return Node.objects.create(author=user)



# Generated by Django 3.2.15 on 2023-02-06 18:36

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("contenttypes", "0002_remove_content_type_name"),
        ("course_flow", "0107_alter_objectpermission_content_type"),
    ]

    operations = [
        migrations.AlterField(
            model_name="objectpermission",
            name="content_type",
            field=models.ForeignKey(
                limit_choices_to={"model__in": ["project", "workflow"]},
                on_delete=django.db.models.deletion.CASCADE,
                to="contenttypes.contenttype",
            ),
        ),
    ]



import json

from django.contrib.auth import get_user_model
from django.core.management.base import BaseCommand

from course_flow.serializers import ActivitySerializer

User = get_user_model()


class Command(BaseCommand):
    def add_arguments(self, parser):
        parser.add_argument("path", type=str)
        parser.add_argument("username", type=str, nargs="?", default=None)

    def handle(self, *args, **kwargs):
        if (
            not User.objects.filter(username=kwargs["username"]).exists()
            and kwargs["username"] is not None
        ):
            return self.stdout.write(
                "No user has the username '%s'" % kwargs["username"]
            )
        json_data = open(kwargs["path"])
        fixtures = json.load(json_data)
        fixtures["author"] = kwargs["username"]
        serializer = ActivitySerializer(data=fixtures)
        if serializer.is_valid():
            serializer.save()
            self.stdout.write("Default strategies have been built")
        else:
            self.stdout.write("Default strategies are not valid")



# Generated by Django 3.2.15 on 2023-01-18 21:09

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("course_flow", "0098_auto_20230118_1413"),
    ]

    operations = [
        migrations.AlterField(
            model_name="courseflowuser",
            name="user",
            field=models.OneToOneField(
                on_delete=django.db.models.deletion.CASCADE,
                primary_key=True,
                related_name="courseflow_user",
                serialize=False,
                to=settings.AUTH_USER_MODEL,
            ),
        ),
    ]



"""
Django settings for course_flow_maker project.

Generated by 'django-admin startproject' using Django 2.2.6.

For more information on this file, see
https://docs.djangoproject.com/en/2.2/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/2.2/ref/settings/
"""

import os

# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))


# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/2.2/howto/deployment/checklist/

DEFAULT_AUTO_FIELD = "django.db.models.AutoField"
SECRET_KEY = "course_flow"
# For LTI tests
PASSWORD_KEY = "course_flow"
LTI_CLIENT_KEY = "course_flow"
LTI_CLIENT_SECRET = "course_flow"

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = False

ALLOWED_HOSTS = ["127.0.0.1", "localhost"]

CHROMEDRIVER_PATH = None

# Application definition

TEACHER_GROUP = "Teacher"

ADMINS = [("John", "john@example.com"), ("Mary", "mary@example.com")]

EMAIL_BACKEND = "django.core.mail.backends.console.EmailBackend"
# EMAIL_BACKEND = 'django.core.mail.backends.smtp.EmailBackend'

DEFAULT_FROM_EMAIL = "noreply@courseflow.org"

CELERY_BROKER_URL = "redis://localhost:6379"
CELERY_RESULT_BACKEND = "redis://localhost:6379"


INSTALLED_APPS = [
    "compressor",
    # "user_feedback.apps.UserFeedbackConfig",
    "course_flow.apps.CourseFlowConfig",
    "rest_framework",
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "daphne",
    "django.contrib.staticfiles",
    "channels",
]


MIDDLEWARE = [
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.locale.LocaleMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
    "csp.middleware.CSPMiddleware",
    "ratelimit.middleware.RatelimitMiddleware",
]

# ASGI
ASGI_APPLICATION = "course_flow.asgi.application"
CHANNEL_LAYERS = {
    "default": {
        "BACKEND": "channels_redis.core.RedisChannelLayer",
        "CONFIG": {"hosts": [("127.0.0.1", 6379)]},
    },
}

# REDIS cache
# CACHES = {
#     "default": {
#         "BACKEND": "django_redis.cache.RedisCache",
#         "LOCATION": "redis://127.0.0.1:6379/1",
#         "OPTIONS": {"CLIENT_CLASS": "django_redis.client.DefaultClient"},
#     }
# }

ROOT_URLCONF = "course_flow.test_urls"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
                "course_flow.context_processors.update_notifications",
            ]
        },
    }
]

CSP_INCLUDE_NONCE_IN = [
    "script-src",
    "style-src",
]
CSP_DEFAULT_SRC = ["'self'", "*.mydalite.org"]
CSP_SCRIPT_SRC = [
    "'self'",
    "*.mydalite.org",
    "d3js.org",
    "ajax.googleapis.com",
    "cdn.polyfill.io",
    "cdn.quilljs.com",
]
CSP_STYLE_SRC = [
    "'self'",
    "*.mydalite.org",
    "ajax.googleapis.com",
    "cdn.quilljs.com",
    "fonts.googleapis.com",
]
CSP_FONT_SRC = [
    "'self'",
    "*.mydalite.org",
    "fonts.gstatic.com",
]

LOGIN_URL = "login"
LOGIN_REDIRECT_URL = "course_flow:home"
LOGOUT_REDIRECT_URL = "login"
COURSE_FLOW_PASSWORD_CHANGE_URL = "login"

ADMINS = [("John", "john@example.com"), ("Mary", "mary@example.com")]

# Database
# https://docs.djangoproject.com/en/2.2/ref/settings/#databases

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": os.path.join(BASE_DIR, "db.sqlite3"),
        "TEST": {"NAME": os.path.join(BASE_DIR, "db_test.sqlite3")},
        "OPTIONS": {"timeout": 20},
    }
}

GRAPH_MODELS = {"all_applications": True, "group_models": True}


# Password validation
# https://docs.djangoproject.com/en/2.2/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {
        "NAME": "django.contrib.auth.password_validation.UserAttribute"
        "SimilarityValidator"
    },
    {"NAME": "django.contrib.auth.password_validation.MinimumLengthValidator"},
    {
        "NAME": "django.contrib.auth.password_validation.CommonPasswordValidator"
    },
    {
        "NAME": "django.contrib.auth.password_validation.NumericPasswordValidator"
    },
]


# Internationalization
# https://docs.djangoproject.com/en/2.2/topics/i18n/

LANGUAGE_CODE = "en-us"

TIME_ZONE = "UTC"

DEFAULT_TIMEZONE = "America/Montreal"

USE_I18N = True

USE_L10N = True

USE_TZ = True


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/2.2/howto/static-files/


STATIC_URL = "/static/"
STATIC_ROOT = os.path.join(BASE_DIR, "server_static/")

STATICFILES_FINDERS = (
    "django.contrib.staticfiles.finders.FileSystemFinder",
    "django.contrib.staticfiles.finders.AppDirectoriesFinder",
    "compressor.finders.CompressorFinder",
)

REST_FRAMEWORK = {
    "DEFAULT_AUTHENTICATION_CLASSES": [
        "rest_framework.authentication.SessionAuthentication"
    ]
}

COURSE_FLOW_RETURN_URL = {"name": "course_flow:home", "title": "myDalite"}

COURSE_FLOW_LTI_ACCESS = True

RATELIMIT_VIEW = "course_flow.views.ratelimited_view"

"""
LOGGING = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "simple": {
            "format": "%(name)s - %(levelname)s - %(asctime)s - %(message)s",
            "datefmt": "%Y-%m-%d %H:%M:%S",
        }
    },
    "handlers": {
        "file": {
            "level": "DEBUG",
            "class": "logging.FileHandler",
            "formatter": "simple",
            "filename": os.path.join(BASE_DIR, "logs"),
        },
        "console": {
            "level": "DEBUG" if DEBUG else "INFO",
            "class": "logging.StreamHandler",
            "formatter": "simple",
            "stream": "ext://sys.stdout",
        },
    },
    "loggers": {
        "django.request": {
            "handlers": ["console"],
            "level": "DEBUG" if DEBUG else "INFO",
            "propagate": True,
        },
        "courseflow": {
            "handlers": ["console", "file"],
            "level": "DEBUG" if DEBUG else "INFO",
        },
    },
}
"""

try:
    from .local_settings import *  # noqa F403

    try:
        INSTALLED_APPS += LOCAL_APPS  # noqa F405
    except NameError:
        pass
except ImportError:
    import warnings

    warnings.warn(
        "File local_settings.py not found. You probably want to add it -- "
        "see README.md."
    )
    pass

# Run this AFTER importing local settings, as this is where debug will usually
# be set to true
if DEBUG:
    INSTALLED_APPS += [
        "debug_toolbar",
    ]

    MIDDLEWARE += [
        "debug_toolbar.middleware.DebugToolbarMiddleware",
    ]



# Generated by Django 2.2.20 on 2021-09-09 21:40

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("course_flow", "0068_workflow_code"),
    ]

    operations = [
        migrations.AddField(
            model_name="workflow",
            name="ponderation_individual",
            field=models.PositiveIntegerField(default=0),
        ),
        migrations.AddField(
            model_name="workflow",
            name="ponderation_practical",
            field=models.PositiveIntegerField(default=0),
        ),
        migrations.AddField(
            model_name="workflow",
            name="ponderation_theory",
            field=models.PositiveIntegerField(default=0),
        ),
        migrations.AddField(
            model_name="workflow",
            name="time_required",
            field=models.CharField(blank=True, max_length=30, null=True),
        ),
        migrations.AddField(
            model_name="workflow",
            name="time_units",
            field=models.PositiveIntegerField(
                choices=[
                    (0, ""),
                    (1, "seconds"),
                    (2, "minutes"),
                    (3, "hours"),
                    (4, "days"),
                    (5, "weeks"),
                    (6, "months"),
                    (7, "yrs"),
                    (8, "credits"),
                ],
                default=0,
            ),
        ),
    ]



# Generated by Django 2.2.20 on 2021-08-27 20:42

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("course_flow", "0065_auto_20210818_2256"),
    ]

    operations = [
        migrations.CreateModel(
            name="CustomTerm",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("term", models.CharField(max_length=50)),
                ("translation", models.CharField(max_length=50)),
            ],
        ),
        migrations.AddField(
            model_name="project",
            name="terminology_dict",
            field=models.ManyToManyField(
                blank=True, to="course_flow.CustomTerm"
            ),
        ),
    ]



import json

from django.contrib.auth.models import Group
from django.http import JsonResponse
from django.urls import reverse


def default_view(request):
    return JsonResponse({"status": "OK"})


# def test_is_owner__known_model(node, settings, users):
#     wrapped_view = decorators.is_owner("nodePk")(default_view)
#     Group.objects.create(name=settings.TEACHER_GROUP)

#     # Logged in and owner -> return view
#     request = HttpRequest()
#     request.method = "POST"
#     request.POST = {"nodePk": json.dumps(node.pk)}
#     request.user = node.author
#     response = wrapped_view(request)

#     assert isinstance(response, JsonResponse)
#     assert response.status_code == 200
#     assert json.loads(response.content)["status"] == "OK"

#     # Logged in and NOT owner -> return 403
#     request.user = users[1]

#     with pytest.raises(PermissionDenied):
#         response = wrapped_view(request)

#     # Not logged in -> return 401
#     request.user = AnonymousUser()
#     response = wrapped_view(request)

#     assert response.status_code == 401
#     assert json.loads(response.content)["login_url"] == settings.LOGIN_URL

#     # Method other than POST -> return 405
#     request.method = "GET"
#     request.user = AnonymousUser()  # Make sure require_POST throws first
#     response = wrapped_view(request)

#     assert response.status_code == 405
#     assert isinstance(response, HttpResponseNotAllowed)

#     # Logged in and null json -> return 400
#     request.method = "POST"
#     request.POST = {}
#     request.user = node.author
#     response = wrapped_view(request)

#     assert response.status_code == 400

#     # Logged in and missing key -> return 400
#     request.method = "POST"
#     request.POST = {"missing": json.dumps("key")}
#     request.user = node.author
#     response = wrapped_view(request)

#     assert response.status_code == 400


# def test_is_owner__unknown_model(node):
#     wrapped_view = decorators.is_owner("unknownModelPk")(default_view)
#     Group.objects.create(name=settings.TEACHER_GROUP)

#     # Logged in -> return 404
#     request = HttpRequest()
#     request.method = "POST"
#     request.POST = {"unknownModelPk": json.dumps("1")}
#     request.user = node.author
#     response = wrapped_view(request)

#     assert response.status_code == 404


# def test_is_owner__False(node):
#     wrapped_view = decorators.is_owner(False)(default_view)
#     Group.objects.create(name=settings.TEACHER_GROUP)

#     # Logged in -> return view
#     request = HttpRequest()
#     request.method = "POST"
#     request.POST = {
#         "objectID": json.dumps(node.pk),
#         "objectType": json.dumps("node"),
#     }
#     request.user = node.author
#     response = wrapped_view(request)

#     assert isinstance(response, JsonResponse)
#     assert response.status_code == 200
#     assert json.loads(response.content)["status"] == "OK"


def test_get_possible_linked_workflows(client, node, settings):
    Group.objects.create(name=settings.TEACHER_GROUP)
    """ Check that ajax_login_required and requre_POST are applied. """

    # Not logged in -> 401
    response = client.post(
        reverse("course_flow:get-possible-linked-workflows"),
        {"nodePk": str(node.pk)},
    )
    assert response.status_code == 401
    assert json.loads(response.content)["login_url"] == settings.LOGIN_URL

    # Not POST -> 405
    response = client.get(
        reverse("course_flow:get-possible-linked-workflows"),
    )
    assert response.status_code == 405



import calendar

from django import template
from django.conf import settings
from django.db.models import Count, Q
from django.urls import reverse
from django.utils import timezone
from django.utils.html import format_html

from course_flow import analytics, models
from course_flow.utils import get_classrooms_for_student

register = template.Library()


@register.simple_tag
def course_flow_return_url():
    if not hasattr(settings, "COURSE_FLOW_RETURN_URL"):
        return reverse("course_flow:home")
    return reverse(
        settings.COURSE_FLOW_RETURN_URL.get("name", "course_flow:home")
    )


@register.simple_tag
def course_flow_return_title():
    if not hasattr(settings, "COURSE_FLOW_RETURN_URL"):
        return "Return Home"
    return settings.COURSE_FLOW_RETURN_URL.get("title", "Return Home")


@register.simple_tag
def course_flow_password_change_url():
    if not hasattr(settings, "COURSE_FLOW_PASSWORD_CHANGE_URL"):
        return reverse("login")
    return reverse(settings.COURSE_FLOW_PASSWORD_CHANGE_URL)


@register.filter
def not_deleted(query):
    return query.filter(deleted=False)


@register.filter
def not_deleted_favourites(query):
    if query is None:
        return None
    return query.filter(
        Q(workflow__deleted=False, workflow__project__deleted=False)
        | Q(project__deleted=False)
    )


@register.filter
def has_group(user, group_name):
    return user.groups.filter(name=group_name).exists()


@register.filter
def get_classrooms(user):
    return get_classrooms_for_student(user)


def get_queryset_by_created_month(queryset, month, year):
    return queryset.filter(
        created_on__month=month, created_on__year=year
    ).count()


def filter_workflow_by_nodes(queryset):
    return queryset.annotate(num_nodes=Count("weeks__nodeweek")).filter(
        num_nodes__gte=3
    )


def filter_project_by_nodes(queryset):
    return queryset.annotate(
        num_nodes=Count("workflows__weekworkflow__week__nodeweek__node")
    ).filter(num_nodes__gte=3)


@register.simple_tag
def get_saltise_admin_workflows():
    # original_month = 1
    # original_year = 2021
    # current_month = timezone.now().month
    # current_year = timezone.now().year
    # projects = models.Project.objects
    # activities = models.Activity.objects
    # courses = models.Course.objects
    # programs = models.Program.objects
    # nodes = models.Node.objects
    # projects_nodes = filter_project_by_nodes(models.Project.objects)
    # activities_nodes = filter_workflow_by_nodes(models.Activity.objects)
    # courses_nodes = filter_workflow_by_nodes(models.Course.objects)
    # programs_nodes = filter_workflow_by_nodes(models.Program.objects)
    # rows = [
    #     "<tr>"
    #     + "<th>Month</th>"
    #     + "<th>Projects</th>"
    #     + "<th>Activities</th>"
    #     + "<th>Courses</th>"
    #     + "<th>Programs</th>"
    #     + "<th>Nodes</th>"
    #     + "</tr>"
    # ]
    # for year in range(original_year, current_year + 1):
    #     first_month = 1
    #     last_month = 12
    #     if year == original_year:
    #         first_month = original_month
    #     if year == current_year:
    #         last_month = current_month
    #     for month in range(first_month, last_month + 1):
    #         rows += [
    #             (
    #                 "<tr>"
    #                 + "<td>{} {}</td>"
    #                 + "<td>{} ({})</td>"
    #                 + "<td>{} ({})</td>"
    #                 + "<td>{} ({})</td>"
    #                 + "<td>{} ({})</td>"
    #                 + "<td>{}</td>"
    #                 + "</tr>"
    #             ).format(
    #                 calendar.month_name[month],
    #                 year,
    #                 get_queryset_by_created_month(projects, month, year),
    #                 get_queryset_by_created_month(projects_nodes, month, year),
    #                 get_queryset_by_created_month(activities, month, year),
    #                 get_queryset_by_created_month(
    #                     activities_nodes, month, year
    #                 ),
    #                 get_queryset_by_created_month(courses, month, year),
    #                 get_queryset_by_created_month(courses_nodes, month, year),
    #                 get_queryset_by_created_month(programs, month, year),
    #                 get_queryset_by_created_month(programs_nodes, month, year),
    #                 get_queryset_by_created_month(nodes, month, year),
    #             )
    #         ]
    # rows += [
    #     (
    #         "<tr>"
    #         + "<td>{}</td>"
    #         + "<td>{} ({})</td>"
    #         + "<td>{} ({})</td>"
    #         + "<td>{} ({})</td>"
    #         + "<td>{} ({})</td>"
    #         + "<td>{}</td>"
    #         + "</tr>"
    #     ).format(
    #         "Total",
    #         projects.all().count(),
    #         projects_nodes.count(),
    #         activities.all().count(),
    #         activities_nodes.count(),
    #         courses.all().count(),
    #         courses_nodes.count(),
    #         programs.all().count(),
    #         programs_nodes.count(),
    #         nodes.all().count(),
    #     )
    # ]
    # table = "\n".join(rows)
    # return format_html("<table>" + table + "</table>")

    return format_html(
        analytics.get_workflow_table().to_html(classes="analytics-workflows")
    )


# def get_annoted_users_for_objects_by_month(type, month, year):
#     if type in ["activity", "course", "program"]:
#         my_filter = {
#             "authored_workflows__" + type + "__created_on__month": month,
#             "authored_workflows__" + type + "__created_on__year": year,
#         }
#         annotation = "authored_workflows__" + type
#     else:
#         my_filter = {
#             "authored_" + type + "__created_on__month": month,
#             "authored_" + type + "__created_on__year": year,
#         }
#         annotation = "authored_" + type
#     result = models.User.objects.filter(**my_filter)
#     return result.annotate(num_authored=Count(annotation))


# def get_users_for_object(queryset):
#     return queryset.filter(num_authored__gte=1).count()


@register.simple_tag
def get_saltise_admin_users():
    # original_month = 1
    # original_year = 2021
    # current_month = timezone.now().month
    # current_year = timezone.now().year
    # rows = [
    #     "<tr>"
    #     + "<th>Month</th>"
    #     + "<th>Project Authors</th>"
    #     + "<th>Activity Authors</th>"
    #     + "<th>Course Authors</th>"
    #     + "<th>Program Authors</th>"
    #     + "<th>Node Authors</th>"
    #     + "</tr>"
    # ]
    # for year in range(original_year, current_year + 1):
    #     first_month = 1
    #     last_month = 12
    #     if year == original_year:
    #         first_month = original_month
    #     if year == current_year:
    #         last_month = current_month
    #     for month in range(first_month, last_month + 1):
    #         annotated_activities = get_annoted_users_for_objects_by_month(
    #             "activity", month, year
    #         )
    #         annotated_courses = get_annoted_users_for_objects_by_month(
    #             "course", month, year
    #         )
    #         annotated_programs = get_annoted_users_for_objects_by_month(
    #             "program", month, year
    #         )
    #         annotated_projects = get_annoted_users_for_objects_by_month(
    #             "projects", month, year
    #         )
    #         annotated_nodes = get_annoted_users_for_objects_by_month(
    #             "nodes", month, year
    #         )
    #         rows += [
    #             (
    #                 "<tr>"
    #                 + "<td>{} {}</td>"
    #                 + "<td>{}</td>"
    #                 + "<td>{}</td>"
    #                 + "<td>{}</td>"
    #                 + "<td>{}</td>"
    #                 + "<td>{}</td>"
    #                 + "</tr>"
    #             ).format(
    #                 calendar.month_name[month],
    #                 year,
    #                 get_users_for_object(annotated_projects),
    #                 get_users_for_object(annotated_activities),
    #                 get_users_for_object(annotated_courses),
    #                 get_users_for_object(annotated_programs),
    #                 get_users_for_object(annotated_nodes),
    #             )
    #         ]
    # rows += [
    #     (
    #         "<tr>"
    #         + "<td>Total</td>"
    #         + "<td>{}</td>"
    #         + "<td>{}</td>"
    #         + "<td>{}</td>"
    #         + "<td>{}</td>"
    #         + "<td>{}</td>"
    #         + "</tr>"
    #     ).format(
    #         get_users_for_object(
    #             models.User.objects.all().annotate(
    #                 num_authored=Count("authored_projects")
    #             )
    #         ),
    #         get_users_for_object(
    #             models.User.objects.all().annotate(
    #                 num_authored=Count("authored_workflows__activity")
    #             )
    #         ),
    #         get_users_for_object(
    #             models.User.objects.all().annotate(
    #                 num_authored=Count("authored_workflows__course")
    #             )
    #         ),
    #         get_users_for_object(
    #             models.User.objects.all().annotate(
    #                 num_authored=Count("authored_workflows__program")
    #             )
    #         ),
    #         get_users_for_object(
    #             models.User.objects.all().annotate(
    #                 num_authored=Count("authored_nodes")
    #             )
    #         ),
    #     )
    # ]
    # table = "\n".join(rows)
    # return format_html("<table>" + table + "</table>")
    return format_html(
        analytics.get_user_table().to_html(classes="analytics-users")
    )


@register.simple_tag
def get_saltise_admin_user_details():
    return format_html(
        analytics.get_user_details_table().to_html(classes="analytics-details")
    )



import calendar
import time

import pandas as pd
from django.core.cache import cache

from course_flow.models import Project, Workflow
from course_flow.serializers import AnalyticsSerializer
from course_flow.utils import benchmark


def month_replace(x):
    return calendar.month_abbr[int(x.string)]


def fix_months(df):
    for i, level in enumerate(df.index.levels):
        if level.name == "Month":
            df.index = df.index.set_levels(
                level.str.replace(r"^\d{1,2}$", month_replace, regex=True),
                level=i,
            )


def get_base_dataframe():

    df = cache.get("COURSEFLOW_ANALYTICS_DATAFRAME", None)
    if df is not None:
        return df

    projects = AnalyticsSerializer(
        Project.objects.exclude(author=None), many=True
    ).data
    workflows = AnalyticsSerializer(
        Workflow.objects.exclude(author=None), many=True
    ).data

    df = pd.DataFrame(projects + workflows)
    created_data = df["created_on"].str.split(" ", expand=True)
    df["Year"] = created_data[0]
    df["Month"] = created_data[1]

    df = df.loc[df["nodes"] >= 3]

    # df["content_rich"] = df["nodes"]>=3
    # df["content_rich"]=pd.Categorical(df["content_rich"])
    # for cat in [True,False]:
    #   if cat not in df["content_rich"].cat.categories:
    #     df["content_rich"] = df["content_rich"].cat.add_categories(cat)
    df["type"] = pd.Categorical(df["type"])
    for cat in ["activity", "course", "program", "project"]:
        if cat not in df["type"].cat.categories:
            df["type"] = df["type"].cat.add_categories(cat)

    df["Institution"] = df["email"].str.rsplit("@", expand=True, n=1)[1]

    # df["Month"] = df["Month"].apply(lambda x: calendar.month_abbr[int(x)])

    cache.set("COURSEFLOW_ANALYTICS_DATAFRAME", df, 600)

    return df


def get_workflow_table(df=None):
    if df is None:
        df = get_base_dataframe()

    pt = pd.pivot_table(
        df,
        values="nodes",
        columns="type",
        index=["Year", "Month"],
        aggfunc="count",
        fill_value=0,
        margins=True,
    )

    pt.fillna(0, inplace=True)
    pt["All"] = pt["All"].astype("int")
    fix_months(pt)
    return pt


def get_user_table(df=None):
    if df is None:
        df = get_base_dataframe()
    df1 = df.copy()
    df1["is_active"] = df1["User"].str.contains("(active)")
    df2 = df1.groupby(["Year", "Month"])
    df3 = df1.groupby(["Year", "Month", "type"])

    user_counts = df2["User"].nunique()
    active_counts = (
        df1.loc[df1["is_active"]].groupby(["Year", "Month"])["User"].nunique()
    )

    type_counts = df3["User"].nunique()
    type_counts = type_counts.unstack()
    type_counts["Total Unique Users"] = user_counts
    type_counts["Total Active Users"] = active_counts

    year_counts = df1.groupby(["Year", "type"])["User"].nunique()
    year_counts_user = df1.groupby(["Year"])["User"].nunique()
    year_counts_active = (
        df1.loc[df1["is_active"]].groupby(["Year"])["User"].nunique()
    )
    year_counts = (
        year_counts.reindex(
            pd.MultiIndex.from_product(
                [level for level in year_counts.index.levels]
                + [["Year Total"]]
            )
        )
        .unstack(1)
        .unstack()
    )
    year_counts["Total Unique Users", "Year Total"] = year_counts_user
    year_counts["Total Active Users", "Year Total"] = year_counts_active

    type_counts = pd.concat([type_counts, year_counts.stack()]).sort_index()
    type_counts.fillna(0, inplace=True)
    type_counts["Total Unique Users"] = type_counts[
        "Total Unique Users"
    ].astype("int")
    type_counts["Total Active Users"] = type_counts[
        "Total Active Users"
    ].astype("int")

    fix_months(type_counts)

    return type_counts


def get_user_details_table(df=None):
    if df is None:
        df = get_base_dataframe()
    df = get_base_dataframe()

    df_sum = (
        df.groupby(["Year", "Month", "Institution", "User", "type"])
        .size()
        .unstack()
    )
    df2 = df.groupby(["Year", "Month", "Institution", "type"]).size().unstack()
    df2 = df2.reindex(
        pd.MultiIndex.from_product(
            [level for level in df2.index.levels] + [["Domain Total"]]
        )
    )
    df3 = df.groupby(["Year", "Month", "type"]).size().unstack()
    df3 = df3.reindex(
        pd.MultiIndex.from_product(
            [level for level in df3.index.levels] + [["Month Total"], [""]]
        )
    )
    df4 = df.groupby(["Year", "type"]).size()
    df4 = df4.reindex(
        pd.MultiIndex.from_product(
            [level for level in df4.index.levels]
            + [["Year Total"], [""], [""]]
        )
    ).unstack(1)
    df5 = df4.sum()

    df_totals = pd.concat([df_sum, df2, df3, df4])

    df_totals = df_totals.loc[~(df_totals.any(axis=1) == 0)]

    def sort_key(index):
        if index.name == "Institution":
            return index.str.replace("Month Total", "zzz")
        return index

    df_totals = df_totals.sort_index(key=sort_key)

    df_totals.loc["Grand Total", :] = df5.values

    df_totals["activity"] = df_totals["activity"].astype("int")
    df_totals["course"] = df_totals["course"].astype("int")
    df_totals["program"] = df_totals["program"].astype("int")
    df_totals["project"] = df_totals["project"].astype("int")

    fix_months(df_totals)

    return df_totals



# Generated by Django 2.2.20 on 2021-07-28 22:38

import django.utils.timezone
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("course_flow", "0062_auto_20210721_1733"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="nodelink",
            name="parent_nodelink",
        ),
        migrations.AlterField(
            model_name="column",
            name="created_on",
            field=models.DateTimeField(default=django.utils.timezone.now),
        ),
        migrations.AlterField(
            model_name="columnworkflow",
            name="added_on",
            field=models.DateTimeField(default=django.utils.timezone.now),
        ),
        migrations.AlterField(
            model_name="node",
            name="created_on",
            field=models.DateTimeField(default=django.utils.timezone.now),
        ),
        migrations.AlterField(
            model_name="nodelink",
            name="created_on",
            field=models.DateTimeField(default=django.utils.timezone.now),
        ),
        migrations.AlterField(
            model_name="nodeweek",
            name="added_on",
            field=models.DateTimeField(default=django.utils.timezone.now),
        ),
        migrations.AlterField(
            model_name="outcome",
            name="created_on",
            field=models.DateTimeField(default=django.utils.timezone.now),
        ),
        migrations.AlterField(
            model_name="outcomehorizontallink",
            name="added_on",
            field=models.DateTimeField(default=django.utils.timezone.now),
        ),
        migrations.AlterField(
            model_name="outcomenode",
            name="added_on",
            field=models.DateTimeField(default=django.utils.timezone.now),
        ),
        migrations.AlterField(
            model_name="outcomeoutcome",
            name="added_on",
            field=models.DateTimeField(default=django.utils.timezone.now),
        ),
        migrations.AlterField(
            model_name="outcomeworkflow",
            name="added_on",
            field=models.DateTimeField(default=django.utils.timezone.now),
        ),
        migrations.AlterField(
            model_name="project",
            name="created_on",
            field=models.DateTimeField(default=django.utils.timezone.now),
        ),
        migrations.AlterField(
            model_name="week",
            name="created_on",
            field=models.DateTimeField(default=django.utils.timezone.now),
        ),
        migrations.AlterField(
            model_name="weekworkflow",
            name="added_on",
            field=models.DateTimeField(default=django.utils.timezone.now),
        ),
        migrations.AlterField(
            model_name="workflow",
            name="created_on",
            field=models.DateTimeField(default=django.utils.timezone.now),
        ),
        migrations.AlterField(
            model_name="workflowproject",
            name="added_on",
            field=models.DateTimeField(default=django.utils.timezone.now),
        ),
    ]



from django.urls import path
from django.views.i18n import JavaScriptCatalog
from ratelimit.decorators import ratelimit
from rest_framework import routers

from . import views

router = routers.SimpleRouter()


app_name = "course_flow"


def course_flow_patterns():
    return [
        path("home/", views.home_view, name="home"),
        path(
            "saltise-admin/",
            views.SALTISEAdminView.as_view(),
            name="saltise-admin",
        ),
        path(
            "admin/saltise-analytics",
            views.SALTISEAnalyticsView.as_view(),
            name="saltise-analytics",
        ),
        path("logout/", views.logout_view, name="logout"),
        path("mylibrary/", views.mylibrary_view, name="my-library"),
        path("mylibrary/get-projects/", views.get_library, name="get-library"),
        path(
            "favourites/get-projects/",
            views.get_favourites,
            name="get-favourites",
        ),
        path("home/get-projects/", views.get_home, name="get-home"),
        path(
            "mylibrary/search-all-objects/",
            views.search_all_objects,
            name="search-all-objects",
        ),
        path(
            "project/get-workflows-for-project/",
            views.get_workflows_for_project,
            name="get-workflows-for-project",
        ),
        # path("myprojects/", views.myprojects_view, name="my-projects"),
        path(
            "myliveprojects/",
            views.my_live_projects_view,
            name="my-live-projects",
        ),
        # path("mytemplates/", views.mytemplates_view, name="my-templates"),
        # path("myshared/", views.myshared_view, name="my-shared"),
        path("myfavourites/", views.myfavourites_view, name="my-favourites"),
        path("explore/", views.ExploreView.as_view(), name="explore"),
        path("import/", views.import_view, name="import"),
        path(
            "project/<int:pk>/",
            views.ProjectDetailView.as_view(),
            name="project-update",
        ),
        path(
            "liveproject/<int:pk>/",
            views.LiveProjectDetailView.as_view(),
            name="live-project-update",
        ),
        path(
            "assignment/<int:pk>/",
            views.AssignmentDetailView.as_view(),
            name="assignment-update",
        ),
        path(
            "project/<int:pk>/comparison",
            views.ProjectComparisonView.as_view(),
            name="project-comparison",
        ),
        path(
            "workflow/<int:pk>/",
            views.WorkflowDetailView.as_view(),
            name="workflow-update",
        ),
        path(
            "workflow/public/<int:pk>/",
            ratelimit(key="ip", method=["GET"], rate="5/m", block=True)(
                views.WorkflowPublicDetailView.as_view()
            ),
            name="workflow-public",
        ),
        path("workflow/updatevalue/", views.update_value, name="update-value"),
        path("workflow/delete-self/", views.delete_self, name="delete-self"),
        path(
            "workflow/delete-self-live/",
            views.delete_self_live,
            name="delete-self-live",
        ),
        path(
            "workflow/restore-self/",
            views.restore_self,
            name="restore-self",
        ),
        path(
            "workflow/delete-self-soft/",
            views.delete_self_soft,
            name="delete-self-soft",
        ),
        path(
            "workflow/update-outcomenode-degree/",
            views.update_outcomenode_degree,
            name="update-outcomenode-degree",
        ),
        path(
            "workflow/duplicate-self/",
            views.duplicate_self,
            name="duplicate-self",
        ),
        path(
            "project/duplicate-workflow/",
            views.duplicate_workflow_ajax,
            name="duplicate-workflow",
        ),
        path(
            "project/duplicate-project/",
            views.duplicate_project_ajax,
            name="duplicate-project",
        ),
        path(
            "project/duplicate-strategy/",
            views.duplicate_strategy_ajax,
            name="duplicate-strategy",
        ),
        path(
            "workflow/insert-sibling/",
            views.insert_sibling,
            name="insert-sibling",
        ),
        path(
            "workflow/insert-child/",
            views.insert_child,
            name="insert-child",
        ),
        path("workflow/inserted-at/", views.inserted_at, name="inserted-at"),
        path(
            "outcome/update-outcomehorizontallink-degree/",
            views.update_outcomehorizontallink_degree,
            name="update-outcomehorizontallink-degree",
        ),
        path("workflow/node/new", views.new_node, name="new-node"),
        path(
            "workflow/outcome/new",
            views.new_outcome_for_workflow,
            name="new-outcome-for-workflow",
        ),
        path("workflow/strategy/add", views.add_strategy, name="add-strategy"),
        path(
            "workflow/strategy/toggle",
            views.week_toggle_strategy,
            name="toggle-strategy",
        ),
        path(
            "project/from-json/",
            views.project_from_json,
            name="project-from-json",
        ),
        path(
            "project/get-disciplines/",
            views.DisciplineListView.as_view(),
            name="get-disciplines",
        ),
        path(
            "favourites/toggle",
            views.toggle_favourite,
            name="toggle-favourite",
        ),
        path("permissions/set", views.set_permission, name="set-permission"),
        path(
            "liveprojectroles/set",
            views.set_liveproject_role,
            name="set-liveproject-role",
        ),
        path(
            "workflow/node/set-linked-workflow/",
            views.set_linked_workflow_ajax,
            name="set-linked-workflow",
        ),
        path(
            "workflow/node-link/new",
            views.new_node_link,
            name="new-node-link",
        ),
        path(
            "workflow/get-possible-linked-workflows/",
            views.get_possible_linked_workflows,
            name="get-possible-linked-workflows",
        ),
        path(
            "workflow/get-possible-added-workflows/",
            views.get_possible_added_workflows,
            name="get-possible-added-workflows",
        ),
        path(
            "workflow/get-workflow-context/",
            views.get_workflow_context,
            name="get-workflow-context",
        ),
        path(
            "workflow/get-target-projects/",
            views.get_target_projects,
            name="get-target-projects",
        ),
        path(
            "workflow/get-workflow-data/",
            views.get_workflow_data,
            name="get-workflow-data",
        ),
        path(
            "workflow/get-workflow-parent-data/",
            views.get_workflow_parent_data,
            name="get-workflow-parent-data",
        ),
        path(
            "workflow/get-workflow-child-data/",
            views.get_workflow_child_data,
            name="get-workflow-child-data",
        ),
        path(
            "workflow/<int:pk>/get-public-workflow-data/",
            views.get_public_workflow_data,
            name="get-public-workflow-data",
        ),
        path(
            "workflow/<int:pk>/get-public-workflow-parent-data/",
            views.get_public_workflow_parent_data,
            name="get-public-workflow-parent-data",
        ),
        path(
            "workflow/<int:pk>/get-public-workflow-child-data/",
            views.get_public_workflow_child_data,
            name="get-public-workflow-child-data",
        ),
        path(
            "project/get-project-data/",
            views.get_project_data,
            name="get-project-data",
        ),
        # path(
        #     "outcome/get-outcome-data/",
        #     views.get_outcome_data,
        #     name="get-outcome-data",
        # ),
        path(
            "project/get-users-for-object/",
            views.get_users_for_object,
            name="get-users-for-object",
        ),
        path(
            "project/get-users-for-liveproject/",
            views.get_users_for_liveproject,
            name="get-users-for-liveproject",
        ),
        path(
            "users/get-user-list/",
            views.get_user_list,
            name="get-user-list",
        ),
        path(
            "user/update/",
            views.UserUpdateView.as_view(),
            name="user-update",
        ),
        path(
            "user/notifications/",
            views.UserNotificationsView.as_view(),
            name="user-notifications",
        ),
        path(
            "project/create/",
            views.ProjectCreateView.as_view(),
            name="project-create",
        ),
        path(
            "program/<int:projectPk>/create/",
            views.ProgramCreateView.as_view(),
            name="program-create",
        ),
        path(
            "course/<int:projectPk>/create/",
            views.CourseCreateView.as_view(),
            name="course-create",
        ),
        path(
            "course-strategy/create/",
            views.CourseStrategyCreateView.as_view(),
            name="course-strategy-create",
        ),
        path(
            "activity/<int:projectPk>/create/",
            views.ActivityCreateView.as_view(),
            name="activity-create",
        ),
        path(
            "activity-strategy/create/",
            views.ActivityStrategyCreateView.as_view(),
            name="activity-strategy-create",
        ),
        path(
            "comments/get/",
            views.get_comments_for_object,
            name="get-comments-for-object",
        ),
        path(
            "terminology/add/",
            views.add_terminology,
            name="add-terminology",
        ),
        path(
            "workflow/updateobjectset/",
            views.update_object_set,
            name="update-object-set",
        ),
        path(
            "comments/add/",
            views.add_comment,
            name="add-comment",
        ),
        path(
            "comments/remove/",
            views.remove_comment,
            name="remove-comment",
        ),
        path(
            "comments/removeall/",
            views.remove_all_comments,
            name="remove-all-comments",
        ),
        path(
            "parentworkflows/get/",
            views.get_parent_workflow_info,
            name="get-parent-workflow-info",
        ),
        path(
            "workflow/<int:pk>/get-public-parent-workflow-info/",
            views.get_public_parent_workflow_info,
            name="get-public-parent-workflow-info",
        ),
        path(
            "exports/get/",
            views.get_export,
            name="get-export",
        ),
        path(
            "imports/import-data/",
            views.import_data,
            name="import-data",
        ),
        path(
            "downloads/saltise/get/",
            views.get_saltise_download,
            name="get-saltise-download",
        ),
        path(
            "downloads/exports/get/",
            views.get_export_download,
            name="get-export-download",
        ),
        path(
            "project/makelive",
            views.make_project_live,
            name="make-project-live",
        ),
        path(
            "liveproject/workflowvisibility",
            views.set_workflow_visibility,
            name="set-workflow-visibility",
        ),
        path(
            "liveproject/getliveprojectdata",
            views.get_live_project_data,
            name="get-live-project-data",
        ),
        path(
            "liveproject/getliveprojectdatastudent",
            views.get_live_project_data_student,
            name="get-live-project-data-student",
        ),
        path(
            "liveproject/getassignmentdata",
            views.get_assignment_data,
            name="get-assignment-data",
        ),
        path(
            "liveproject/getassignmentdatastudent",
            views.get_assignment_data_student,
            name="get-assignment-data-student",
        ),
        path(
            "liveproject/getworkflownodes",
            views.get_workflow_nodes,
            name="get-workflow-nodes",
        ),
        path(
            "liveproject/createliveassignment",
            views.create_live_assignment,
            name="create-live-assignment",
        ),
        path(
            "liveproject/updatevalue/",
            views.update_liveproject_value,
            name="update-liveproject-value",
        ),
        path(
            "assignment/addusers/",
            views.add_users_to_assignment,
            name="add-users-to-assignment",
        ),
        path(
            "assignment/setcompletion/",
            views.set_assignment_completion,
            name="set-assignment-completion",
        ),
        path(
            "assignment/getassignmentsfornode/",
            views.get_assignments_for_node,
            name="get-assignments-for-node",
        ),
        path(
            "liveproject/register/student/<project_hash>/",
            views.register_as_student,
            name="register-as-student",
        ),
        path(
            "user/select-notifications/",
            views.select_notifications,
            name="select-notifications",
        ),
        path(
            "user/mark-all-as-read/",
            views.mark_all_as_read,
            name="mark-all-as-read",
        ),
        path(
            "jsi18n/", JavaScriptCatalog.as_view(), name="javascript-catalog"
        ),
    ] + router.urls


urlpatterns = course_flow_patterns()



# Generated by Django 2.2.25 on 2022-05-25 22:36

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("course_flow", "0084_week_is_dropped"),
    ]

    operations = [
        migrations.AddField(
            model_name="nodelink",
            name="text_position",
            field=models.PositiveSmallIntegerField(default=20),
        ),
    ]



# Generated by Django 3.2.15 on 2023-02-05 16:27

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("course_flow", "0103_auto_20230205_1626"),
    ]

    operations = [
        migrations.RenameField(
            model_name="workflow",
            old_name="author_temp",
            new_name="author",
        ),
    ]



from io import BytesIO

import pandas as pd
from django.db.models import Q
from django.utils.translation import gettext as _

from course_flow import analytics

from .models import (
    Course,
    Node,
    OutcomeNode,
    OutcomeWorkflow,
    Program,
    Week,
    WeekWorkflow,
)
from .serializers import (
    NodeExportSerializer,
    NodeExportSerializerWithTime,
    OutcomeExportSerializer,
    WeekExportSerializer,
    WorkflowExportSerializer,
)
from .utils import (
    get_all_outcomes_ordered_filtered,
    get_all_outcomes_ordered_for_outcome,
    get_alphanum,
    get_base_outcomes_ordered_filtered,
    get_outcomenodes,
    get_parent_nodes_for_workflow,
    get_unique_outcomehorizontallinks,
)


def concat_line(df, dict):
    return pd.concat([df, pd.DataFrame([dict])])


def concat_df(df1, df2):
    return pd.concat([df1, df2])


def get_str(obj, key):
    s = obj.get(key, "")
    return "" if s is None else s


def stringify(value):
    if value is None:
        return ""
    else:
        return str(value)


def allowed_sets_Q(allowed_sets):
    return Q(sets__in=allowed_sets) | Q(sets=None)


def check_allowed_sets(obj, allowed_sets):
    if obj.sets.all().count() == 0:
        return True
    if (
        obj.sets.filter(
            id__in=allowed_sets.values_list("id", flat=True)
        ).count()
        > 0
    ):
        return True
    return False


def get_framework_line_for_outcome(outcome, columns, allowed_sets):
    outcome_serialized = OutcomeExportSerializer(outcome).data
    sub_outcomes = get_all_outcomes_ordered_for_outcome(outcome)
    sub_outcomes_serialized = OutcomeExportSerializer(
        sub_outcomes[1:], many=True
    ).data
    sub_outcomes_entry = "\n".join(
        [
            get_str(sub, "code") + " - " + get_str(sub, "title")
            for sub in sub_outcomes_serialized
        ]
    )
    outcomes_horizontal = [
        och.parent_outcome
        for och in get_unique_outcomehorizontallinks(outcome)
    ]
    outcomes_horizontal_serialized = OutcomeExportSerializer(
        filter(
            lambda och: check_allowed_sets(och, allowed_sets),
            outcomes_horizontal,
        ),
        many=True,
    ).data
    outcomes_horizontal_entry = "\n".join(
        [get_str(och, "code") for och in outcomes_horizontal_serialized]
    )
    dict_data = {
        "0": get_str(outcome_serialized, "code")
        + " - "
        + get_str(outcome_serialized, "title"),
        "1": sub_outcomes_entry,
        "2": outcomes_horizontal_entry,
    }
    for i, column in enumerate(columns):
        nodes = (
            Node.objects.filter(
                outcomenode__outcome__in=sub_outcomes,
                column=column,
                deleted=False,
            )
            .filter(allowed_sets_Q(allowed_sets))
            .distinct()
        )
        nodes_serialized = NodeExportSerializer(nodes, many=True).data
        dict_data[str(3 + i)] = "\n".join(
            [node["title"] for node in nodes_serialized]
        )
    return dict_data


def get_course_framework(workflow, allowed_sets):
    workflow_serialized = WorkflowExportSerializer(workflow).data
    num_columns = workflow.columns.all().count()
    # df_columns = max(6, 3 + num_columns)
    df = pd.DataFrame(columns=[str(i) for i in range(num_columns)])
    df = concat_line(
        df,
        {
            "0": _("Course Title"),
            "1": workflow_serialized["title"],
            "2": _("Ponderation Theory/Practical/Individual"),
            "3": str(workflow.ponderation_theory)
            + "/"
            + str(workflow.ponderation_practical)
            + "/"
            + str(workflow.ponderation_individual),
        },
    )
    df = concat_line(
        df,
        {
            "0": _("Course Description"),
            "1": workflow_serialized["description"],
        },
    )
    df = concat_line(
        df,
        {
            "0": _("Course Code"),
            "1": workflow.code,
            "2": _("Hours"),
            "3": str(
                workflow.time_general_hours + workflow.time_specific_hours
            ),
            "4": _("Time"),
            "5": stringify(workflow.time_required)
            + " "
            + workflow.get_time_units_display(),
        },
    )
    df = concat_line(df, {"0": _("Ministerial Competencies")})
    df = concat_line(df, {"0": _("Competency"), "1": _("Title")})
    nodes = (
        get_parent_nodes_for_workflow(workflow)
        .filter(allowed_sets_Q(allowed_sets))
        .distinct()
    )
    parent_outcomes = []
    for node in nodes:
        outcomenodes = get_outcomenodes(node)
        outcomes_unfiltered = [ocn.outcome for ocn in outcomenodes]
        outcomes_filtered = filter(
            lambda oc: check_allowed_sets(oc, allowed_sets),
            outcomes_unfiltered,
        )
        parent_outcomes += OutcomeExportSerializer(
            outcomes_filtered, many=True
        ).data
    a = [get_str(outcome, "code") for outcome in parent_outcomes]
    b = [get_str(outcome, "title") for outcome in parent_outcomes]
    df = pd.concat(
        [
            df,
            pd.DataFrame({"0": a, "1": b}).sort_values(
                by="0",
            ),
        ]
    )
    parent_outcome_length = len(parent_outcomes)
    if len(nodes) > 0:
        df = concat_line(
            df,
            {
                "0": _("Term"),
                "1": WeekWorkflow.objects.get(
                    week__nodes=nodes[0]
                ).get_display_rank()
                + 1,
            },
        )
        prereqs = (
            Node.objects.filter(
                outgoing_links__target_node__in=nodes,
                deleted=False,
            )
            .filter(allowed_sets_Q(allowed_sets))
            .distinct()
        )
        postreqs = (
            Node.objects.filter(
                incoming_links__source_node__in=nodes,
                deleted=False,
            )
            .filter(allowed_sets_Q(allowed_sets))
            .distinct()
        )
        if len(prereqs) > 0:
            df = concat_line(
                df,
                {
                    "0": _("Prerequisites"),
                    "1": ", ".join(
                        [
                            req["title"]
                            for req in NodeExportSerializer(
                                prereqs, many=True
                            ).data
                        ]
                    ),
                },
            )
        if len(postreqs) > 0:
            df = concat_line(
                df,
                {
                    "0": _("Required For"),
                    "1": ", ".join(
                        [
                            req["title"]
                            for req in NodeExportSerializer(
                                postreqs, many=True
                            ).data
                        ]
                    ),
                },
            )
    headers = {
        "0": _("Course Outcome"),
        "1": _("Sub-Outcomes"),
        "2": _("Competencies"),
    }
    columns = workflow.columns.order_by("columnworkflow__rank").all()
    for i, column in enumerate(columns):
        headers[str(3 + i)] = column.get_display_title()
    df = concat_line(df, headers)
    for outcome in workflow.outcomes.filter(deleted=False).filter(
        allowed_sets_Q(allowed_sets)
    ):
        df = concat_line(
            df, get_framework_line_for_outcome(outcome, columns, allowed_sets)
        )
    return df, parent_outcome_length


def get_workflow_outcomes_table(workflow, allowed_sets):
    outcomes = get_all_outcomes_ordered_filtered(
        workflow, allowed_sets_Q(allowed_sets)
    )
    data = OutcomeExportSerializer(outcomes, many=True).data
    df = pd.DataFrame(
        data, columns=["code", "title", "description", "id", "depth"]
    )
    pd.set_option("display.max_colwidth", None)
    return df


def get_outcomes_export(
    model_object, object_type, export_format, allowed_sets
):
    if object_type == "project":
        workflows = list(model_object.workflows.filter(deleted=False))
    else:
        workflows = [model_object]
    with BytesIO() as b:
        if export_format == "excel":
            with pd.ExcelWriter(b, engine="xlsxwriter") as writer:
                for workflow in workflows:
                    df = get_workflow_outcomes_table(workflow, allowed_sets)
                    sheet_name = (
                        get_alphanum(workflow.title) + "_" + str(workflow.pk)
                    )[:30]
                    df.to_excel(
                        writer,
                        sheet_name=sheet_name,
                        index=False,
                    )
        elif export_format == "csv":
            df = pd.DataFrame(
                {}, columns=["code", "title", "description", "id", "depth"]
            )
            workflows_serialized = WorkflowExportSerializer(
                workflows, many=True
            ).data
            for i, workflow in enumerate(workflows):
                df = concat_line(
                    df, {"title": workflows_serialized[i]["title"]}
                )
                df = concat_df(
                    df, get_workflow_outcomes_table(workflow, allowed_sets)
                )
                df = concat_line(df, {"title": ""})
            df.to_csv(path_or_buf=b, sep=",", index=False)
        return b.getvalue()


def get_course_frameworks_export(
    model_object, object_type, export_format, allowed_sets
):
    if object_type == "project":
        workflows = list(
            Course.objects.filter(project=model_object, deleted=False)
        )
    else:
        workflows = [model_object]
    with BytesIO() as b:
        if export_format == "excel":
            with pd.ExcelWriter(b, engine="xlsxwriter") as writer:
                workbook = writer.book
                header_format = workbook.add_format({"bg_color": "#b5fbbb"})
                bold_format = workbook.add_format(
                    {"bold": True, "bg_color": "#04BA74", "color": "white"}
                )
                wrap_format = workbook.add_format()
                wrap_format.set_text_wrap()
                wrap_format.set_align("top")

                for workflow in workflows:
                    df, parent_outcome_length = get_course_framework(
                        workflow, allowed_sets
                    )
                    sheet_name = (
                        get_alphanum(workflow.title) + "_" + str(workflow.pk)
                    )[:30]
                    df.to_excel(
                        writer,
                        sheet_name=sheet_name,
                        index=False,
                    )
                    worksheet = writer.sheets[sheet_name]
                    worksheet.set_column(0, 0, 20, wrap_format)
                    worksheet.set_column(1, 1, 40, wrap_format)
                    worksheet.set_column(2, 99, 25, wrap_format)

                    worksheet.set_row(
                        5 + parent_outcome_length + 4, None, bold_format
                    )
                    bold_cells = [
                        [1, 0],
                        [1, 2],
                        [2, 0],
                        [3, 0],
                        [3, 2],
                        [3, 4],
                        [4, 0],
                        [5, 0],
                        [5, 1],
                        [5 + parent_outcome_length + 1, 0],
                        [5 + parent_outcome_length + 2, 0],
                        [5 + parent_outcome_length + 3, 0],
                    ]
                    for cell in bold_cells:
                        worksheet.conditional_format(
                            cell[0],
                            cell[1],
                            cell[0],
                            cell[1],
                            {"type": "no_errors", "format": bold_format},
                        )

                    worksheet.merge_range(2, 1, 2, 5, None)
                    # try:
                    #     parent_outcome_length = df.get("parent_outcome_length",0)
                    #     for i in range(parent_outcome_length):
                    #         worksheet.merge_range(6+i,2,6+i,5)
                    #         worksheet.write()
        elif export_format == "csv":
            df = pd.DataFrame({})
            workflows_serialized = WorkflowExportSerializer(
                workflows, many=True
            ).data
            for i, workflow in enumerate(workflows):
                df = concat_line(df, {"0": workflows_serialized[i]["title"]})
                df = concat_df(
                    df, get_course_framework(workflow, allowed_sets)[0]
                )
                df = concat_line(df, {"0": ""})
            df.to_csv(path_or_buf=b, sep=",", index=False)
        return b.getvalue()


def get_program_matrix_export(
    model_object, object_type, export_format, allowed_sets
):
    if object_type == "project":
        workflows = list(
            Program.objects.filter(project=model_object, deleted=False)
        )
    else:
        workflows = [model_object]
    with BytesIO() as b:
        if export_format == "excel":
            with pd.ExcelWriter(b, engine="xlsxwriter") as writer:
                for workflow in workflows:
                    df = get_program_matrix(workflow, True, allowed_sets)
                    sheet_name = (
                        get_alphanum(workflow.title) + "_" + str(workflow.pk)
                    )[:30]
                    df.to_excel(
                        writer,
                        sheet_name=sheet_name,
                        index=False,
                    )
        elif export_format == "csv":
            df = pd.DataFrame({})
            workflows_serialized = WorkflowExportSerializer(
                workflows, many=True
            ).data
            for i, workflow in enumerate(workflows):
                df = concat_line(df, {"0": workflows_serialized[i]["title"]})
                df = concat_df(
                    df, get_program_matrix(workflow, True, allowed_sets)
                )
                df = concat_line(df, {"0": ""})
            df.to_csv(path_or_buf=b, sep=",", index=False)
        return b.getvalue()


def get_sobec_export(model_object, object_type, export_format, allowed_sets):
    if object_type == "project":
        workflows = list(
            Program.objects.filter(project=model_object, deleted=False)
        )
    else:
        workflows = [model_object]
    with BytesIO() as b:
        if export_format == "excel":
            with pd.ExcelWriter(b, engine="xlsxwriter") as writer:
                workbook = writer.book
                header_format = workbook.add_format({"bg_color": "#b5fbbb"})
                bold_format = workbook.add_format(
                    {"bold": True, "bg_color": "#04BA74", "color": "white"}
                )
                wrap_format = workbook.add_format()
                wrap_format.set_text_wrap()
                wrap_format.set_align("top")
                for workflow in workflows:
                    df = get_sobec(workflow, allowed_sets)
                    sheet_name = (
                        get_alphanum(workflow.title) + "_" + str(workflow.pk)
                    )[:30]
                    df.to_excel(
                        writer,
                        sheet_name=sheet_name,
                        index=False,
                    )
                    worksheet = writer.sheets[sheet_name]
                    worksheet.set_column(0, 0, 30, wrap_format)
                    worksheet.set_column(1, 1, 40, wrap_format)
                    worksheet.set_column(2, 2, 50, wrap_format)
                    worksheet.set_column(3, 3, 10, wrap_format)
                    worksheet.set_column(4, 4, 10, wrap_format)
                    worksheet.set_column(5, 5, 10, wrap_format)
                    worksheet.set_column(6, 6, 20, wrap_format)
                    worksheet.set_row(0, None, bold_format)
        elif export_format == "csv":
            df = pd.DataFrame({})
            workflows_serialized = WorkflowExportSerializer(
                workflows, many=True
            ).data
            for i, workflow in enumerate(workflows):
                df = concat_line(df, {"0": workflows_serialized[i]["title"]})
                df = concat_df(df, get_sobec(workflow, allowed_sets))
                df = concat_line(df, {"0": ""})
            df.to_csv(path_or_buf=b, sep=",", index=False)
        return b.getvalue()


def get_nodes_export(model_object, object_type, export_format, allowed_sets):
    if object_type == "project":
        workflows = list(model_object.workflows.filter(deleted=False))
    else:
        workflows = [model_object]
    with BytesIO() as b:
        if export_format == "excel":
            with pd.ExcelWriter(b, engine="xlsxwriter") as writer:
                for workflow in workflows:
                    df = get_workflow_nodes_table(workflow, allowed_sets)
                    sheet_name = (
                        get_alphanum(workflow.title) + "_" + str(workflow.pk)
                    )[:30]
                    df.to_excel(
                        writer,
                        sheet_name=sheet_name,
                        index=False,
                    )
        elif export_format == "csv":
            df = pd.DataFrame(
                {},
                columns=["type", "title", "description", "column_order", "id"],
            )
            workflows_serialized = WorkflowExportSerializer(
                workflows, many=True
            ).data
            for i, workflow in enumerate(workflows):
                df = concat_line(
                    df, {"title": workflows_serialized[i]["title"]}
                )
                df = concat_df(
                    df, get_workflow_nodes_table(workflow, allowed_sets)
                )
                df = concat_line(df, {"title": ""})
            df.to_csv(path_or_buf=b, sep=",", index=False)
        return b.getvalue()


def get_matrix_row_header(row):
    if row["type"] == "node":
        node = row["object"]
        return NodeExportSerializer(node).data["title"]
    else:
        week = row["object"]
        return WeekExportSerializer(week).data["title"]


def evaluate_outcome(node, outcomes):
    outcome_ids = list(node.outcomes.all().values_list("id", flat=True))
    if outcomes[0].id in outcome_ids:
        outcomenode = OutcomeNode.objects.get(outcome=outcomes[0], node=node)
        degree = outcomenode.degree
        if degree == 1:
            return "X"
        value = ""
        if degree & 2:
            value += "I"
        if degree & 4:
            value += "D"
        if degree & 8:
            value += "A"
        return value
    else:
        for sub_outcome in outcomes[1:]:
            if sub_outcome.id in outcome_ids:
                return "(X)"


def get_matrix_entry(row, outcomes):
    if row["type"] == "node":
        node = row["object"]
        return evaluate_outcome(node, outcomes)
    else:
        # week = row["object"]
        return ""


def get_matrix_sum_line(rows, fn):
    total = 0
    values = []
    for row in reversed(rows):
        if row["type"] == "node":
            value = fn(row["object"])
            values += [value]
            if value is None or value == "":
                value = 0
            try:
                this_value = float(value)
            except Exception:
                this_value = 0
            total += this_value
        else:
            values += [total]
            total = 0

    return list(reversed(values))


def get_program_matrix(workflow, simple, allowed_sets):
    if simple:
        outcomes = filter(
            lambda oc: check_allowed_sets(oc, allowed_sets),
            [
                outcomeworkflow.outcome
                for outcomeworkflow in OutcomeWorkflow.objects.filter(
                    workflow=workflow, outcome__deleted=False
                )
            ],
        )
    else:
        outcomes = get_all_outcomes_ordered_filtered(
            workflow, allowed_sets_Q(allowed_sets)
        )

    rows = []
    for weekworkflow in WeekWorkflow.objects.filter(
        workflow=workflow, week__deleted=False
    ).order_by("rank"):
        week = weekworkflow.week
        # Add a line for the term
        rows += [{"type": "week", "object": week}]
        for node in (
            Node.objects.filter(allowed_sets_Q(allowed_sets))
            .filter(week=week, deleted=False)
            .order_by("nodeweek__rank")
        ):
            rows += [{"type": "node", "object": node}]

    # headers
    col = 0
    data = {}
    data[str(col)] = ["", ""] + [get_matrix_row_header(row) for row in rows]
    col += 1

    for outcome in outcomes:
        outcome_data = OutcomeExportSerializer(outcome).data
        all_outcomes = get_all_outcomes_ordered_for_outcome(outcome)
        data[str(col)] = [outcome_data["code"], outcome_data["title"]] + [
            get_matrix_entry(row, all_outcomes) for row in rows
        ]
        col += 1

    data[str(col)] = ["" for x in data["0"]]
    col += 1

    data[str(col)] = ["", _("Specific Education")] + get_matrix_sum_line(
        rows,
        lambda x: x.linked_workflow.time_specific_hours
        if (x.linked_workflow is not None and x.represents_workflow)
        else x.time_specific_hours,
    )
    col += 1

    data[str(col)] = ["", _("General Education")] + get_matrix_sum_line(
        rows,
        lambda x: x.linked_workflow.time_general_hours
        if (x.linked_workflow is not None and x.represents_workflow)
        else x.time_general_hours,
    )
    col += 1

    data[str(col)] = ["", _("Total Hours")] + get_matrix_sum_line(
        rows,
        lambda x: x.linked_workflow.time_general_hours
        + x.linked_workflow.time_specific_hours
        if (x.linked_workflow is not None and x.represents_workflow)
        else x.time_general_hours + x.time_specific_hours,
    )
    col += 1

    data[str(col)] = ["" for x in data["0"]]
    col += 1

    data[str(col)] = ["", _("Theory")] + get_matrix_sum_line(
        rows,
        lambda x: x.linked_workflow.ponderation_theory
        if (x.linked_workflow is not None and x.represents_workflow)
        else x.ponderation_theory,
    )
    col += 1

    data[str(col)] = ["", _("Practical")] + get_matrix_sum_line(
        rows,
        lambda x: x.linked_workflow.ponderation_practical
        if (x.linked_workflow is not None and x.represents_workflow)
        else x.ponderation_practical,
    )
    col += 1

    data[str(col)] = ["", _("Practical")] + get_matrix_sum_line(
        rows,
        lambda x: x.linked_workflow.ponderation_individual
        if (x.linked_workflow is not None and x.represents_workflow)
        else x.ponderation_individual,
    )
    col += 1

    data[str(col)] = ["", _("Total")] + get_matrix_sum_line(
        rows,
        lambda x: x.linked_workflow.ponderation_individual
        + x.linked_workflow.ponderation_theory
        + x.linked_workflow.ponderation_practical
        if (x.linked_workflow is not None and x.represents_workflow)
        else x.ponderation_individual
        + x.ponderation_theory
        + x.ponderation_practical,
    )
    col += 1

    data[str(col)] = ["" for x in data["0"]]
    col += 1

    data[str(col)] = ["", _("Credits")] + get_matrix_sum_line(
        rows,
        lambda x: x.linked_workflow.time_required
        if (x.linked_workflow is not None and x.represents_workflow)
        else x.time_required,
    )
    col += 1

    df = pd.DataFrame(data, columns=[str(x) for x in range(col)])
    return df


def get_workflow_nodes_table(workflow, allowed_sets):
    entries = []
    for week in Week.objects.filter(workflow=workflow, deleted=False).order_by(
        "weekworkflow__rank"
    ):
        entries += [WeekExportSerializer(week).data]
        entries += NodeExportSerializer(
            Node.objects.filter(week=week, deleted=False)
            .filter(allowed_sets_Q(allowed_sets))
            .order_by("nodeweek__rank"),
            many=True,
        ).data
    df = pd.DataFrame(
        entries, columns=["type", "title", "description", "column_order", "id"]
    )
    pd.set_option("display.max_colwidth", None)
    return df


def get_sobec_outcome(workflow, outcome, allowed_sets):
    nodes = (
        Node.objects.filter(week__workflow=workflow)
        .filter(deleted=False)
        .filter(allowed_sets_Q(allowed_sets))
        .filter(
            Q(outcomes=outcome)
            | Q(
                outcomes__parent_outcomes=outcome,
                outcomes__parent_outcomes__deleted=False,
            )
            | Q(
                outcomes__parent_outcomes__parent_outcomes=outcome,
                outcomes__parent_outcomes__parent_outcomes__deleted=False,
            )
        )
        .distinct()
    )
    header = {
        "comp_code": outcome.code,
        "code": f"Pass X of the following courses ({nodes.count()})",
    }
    nodes_serialized = NodeExportSerializerWithTime(nodes, many=True).data
    return [header] + nodes_serialized


def get_sobec(workflow, allowed_sets):
    outcomes = get_base_outcomes_ordered_filtered(
        workflow, allowed_sets_Q(allowed_sets)
    )
    data = []
    for outcome in outcomes:
        data += get_sobec_outcome(workflow, outcome, allowed_sets)

    df = pd.DataFrame(
        data,
        columns=[
            "comp_code",
            "code",
            "title",
            "ponderation_theory",
            "ponderation_practical",
            "ponderation_individual",
            "time_required",
        ],
    )
    df.rename(
        columns={
            "comp_code": _("Competency Code"),
            "code": _("Course Code"),
            "title": _("Title"),
            "ponderation_theory": _("Theory"),
            "ponderation_practical": _("Practical"),
            "ponderation_individual": _("Individual"),
            "time_required": _("Credits"),
        },
        inplace=True,
    )
    pd.set_option("display.max_colwidth", None)
    return df


def get_saltise_analytics():

    df = analytics.get_base_dataframe()

    with BytesIO() as b:
        with pd.ExcelWriter(b, engine="xlsxwriter") as writer:
            analytics.get_workflow_table(df).to_excel(
                writer,
                sheet_name="Workflow Overview",
            )
            analytics.get_user_table(df).to_excel(
                writer,
                sheet_name="User Overview",
            )
            analytics.get_user_details_table(df).to_excel(
                writer,
                sheet_name="User Details",
            )

        return b.getvalue()



from django.urls import re_path

from course_flow import consumers

websocket_urlpatterns = [
    re_path(
        r"ws/update/(?P<workflowPk>[0-9]+)/$",
        consumers.WorkflowUpdateConsumer.as_asgi(),
    ),
]



from django.conf import settings
from django.contrib.auth.models import Group, User

from course_flow.utils import get_model_from_str


def make_object(model_key, author=None):
    if model_key == "column":
        return get_model_from_str(model_key).objects.create(
            title="test" + model_key + "title", author=author
        )
    else:
        return get_model_from_str(model_key).objects.create(
            title="test" + model_key + "title",
            description="test" + model_key + "description",
            author=author,
        )


def login(test_case):
    user = User.objects.create(username="testuser1")
    user.set_password("testpass1")
    user.save()
    teacher_group, _ = Group.objects.get_or_create(name=settings.TEACHER_GROUP)
    user.groups.add(teacher_group)
    logged_in = test_case.client.login(
        username="testuser1", password="testpass1"
    )
    test_case.assertTrue(logged_in)
    return user


def login_student(test_case):
    user = User.objects.create(username="testuser1")
    user.set_password("testpass1")
    user.save()
    logged_in = test_case.client.login(
        username="testuser1", password="testpass1"
    )
    test_case.assertTrue(logged_in)
    return user


def get_author(integer="2"):
    author = User.objects.create(username="testuser" + integer)
    author.set_password("testpass" + integer)
    author.save()
    teacher_group, _ = Group.objects.get_or_create(name=settings.TEACHER_GROUP)
    author.groups.add(teacher_group)
    return author


def check_order(test_case, object_links):
    sorted_links = object_links.order_by("rank")
    for i, link in enumerate(sorted_links):
        test_case.assertEqual(link.rank, i)



import hashlib

from django.conf import settings


def generate_password(username: str) -> str:
    """
    Generates the password created from the `username`. This will always be the
    same as long as `PASSWORD_KEY` doesn't change.
    Parameters
    ----------
    username : str
        Username used to create the password
    Returns
    -------
    str
        Password as a hashed string of length 64
    """
    return hashlib.sha3_256(
        f"{username}.{settings.PASSWORD_KEY}".encode()
    ).hexdigest()



# Generated by Django 2.2.25 on 2022-07-04 18:28

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("course_flow", "0086_workflow_static_view"),
    ]

    operations = [
        migrations.RenameField(
            model_name="workflow",
            old_name="static_view",
            new_name="public_view",
        ),
    ]



# Generated by Django 3.2.14 on 2022-09-30 15:27

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("course_flow", "0088_liveproject"),
    ]

    operations = [
        migrations.AddField(
            model_name="liveproject",
            name="default_all_workflows_visible",
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name="liveproject",
            name="visible_workflows",
            field=models.ManyToManyField(
                blank=True, to="course_flow.Workflow"
            ),
        ),
        migrations.AlterField(
            model_name="objectpermission",
            name="permission_type",
            field=models.PositiveIntegerField(
                choices=[
                    (0, "None"),
                    (1, "View"),
                    (2, "Edit"),
                    (3, "Comment"),
                    (4, "Student"),
                ],
                default=0,
            ),
        ),
    ]



# Generated by Django 3.2.15 on 2022-11-14 19:09

import django.utils.timezone
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("course_flow", "0094_liveassignment_userassignment"),
    ]

    operations = [
        migrations.CreateModel(
            name="UpdateNotification",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "title",
                    models.CharField(blank=True, max_length=200, null=True),
                ),
                (
                    "created_on",
                    models.DateTimeField(default=django.utils.timezone.now),
                ),
            ],
        ),
    ]



# Generated by Django 2.2.25 on 2022-03-08 19:56

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("course_flow", "0079_auto_20220228_1845"),
    ]

    operations = [
        migrations.AddField(
            model_name="node",
            name="ponderation_individual",
            field=models.PositiveIntegerField(default=0, null=True),
        ),
        migrations.AddField(
            model_name="node",
            name="ponderation_practical",
            field=models.PositiveIntegerField(default=0, null=True),
        ),
        migrations.AddField(
            model_name="node",
            name="ponderation_theory",
            field=models.PositiveIntegerField(default=0, null=True),
        ),
        migrations.AddField(
            model_name="node",
            name="time_general_hours",
            field=models.PositiveIntegerField(default=0, null=True),
        ),
        migrations.AddField(
            model_name="node",
            name="time_specific_hours",
            field=models.PositiveIntegerField(default=0, null=True),
        ),
    ]
